<?php
//
// bodycache.php provides caching mechanism for PukiWiki's message body.
// 
// Author  : ryu1k
// mailto  : ryu1kkb ATMARK gmail.com
// URL     : http://debian.fam.cx/experimental/index.php?namazuTamer%2FPukiWiki%2Fbodycache
// License : GPL v2 or (at your option) any later version

// 
// bodycache cache file format:
// 
// Rev. 2: Modified by Jasmin ( http://jasmin.sakura.ne.jp/ )
//  line 0 : bodycache signature : "__PUKIWIKI_BODYCACHE__"
//  line 1 : revision in int : "Rev: NUM"
//  line 2 : title : "Title: Title"
//  line 3 : page name : "Page: Page"
//  line 4 : number of items of footnote by ''((note))' command. : "Footnotes: NUM"
//  lines  : footnote element. "index{int}:footnote{encode_base64(gzcompress(raw_note))}"
//  rest   : body in raw html.
// 
// Rev. 1:
//  line 0 : bodycache signature : "__PUKIWIKI_BODYCACHE__"
//  line 1 : revision in int : "Rev: NUM"
//  line 2 : number of items of footnote by ''((note))' command. : "Footnotes: NUM"
//  lines  : footnote element. "index{int}:footnote{encode_base64(gzcompress(raw_note))}"
//  rest   : body in raw html.
// 
// Rev. 0:
//  0byte- : body
// 

$bodycache_rev = 2;
$bodycache_db_signature = "__PUKIWIKI_BODYCACHE__";
$bodycache_status = "suspended";
$bodycache_del_depth = 2;

function bcdump( $var )
{
    return;

      // $fp = fopen("/tmp/bcdump.txt", "a");
      // fwrite($fp, var_export($var, true) . "\n" );
      // fclose($fp);
}

//! generate bodycache signature. (for skin.)
function bodycache_signature_gen()
{
    global $bodycache_rev;
    global $bodycache_status;
    return 
        'PukiWiki ' .         
        bodycache_get_distribute_anchor('bodycache patch') . 
        ' rev. ' . $bodycache_rev . ' applied. ' .
        'State: ' . $bodycache_status . "."; 
}

function bodycache_get_distribute_anchor( $key )
{
    return '<a href="http://debian.fam.cx/experimental/index.php?namazuTamer%2FPukiWiki%2Fbodycache">' . $key . "</a>"
           .' modified by <a href="http://jasmin.sakura.ne.jp/">Jasmin</a>';
}


function bodycache_pickle( $instr )
{
    return base64_encode( gzcompress( $instr ) );
}


function bodycache_unpickle( $pstr )
{
    return gzuncompress( base64_decode( $pstr ) );
}


function isDynamicPlugin( $name ) 
{
    global $bodycache_disable_plugins;

    $name = str_replace( "\r\n", "", $name);
    $name = str_replace( "\n", "", $name);

    return in_array( strtolower($name), $bodycache_disable_plugins );
}

//! check the given source contain dynamic tag or not.
// if no special tag found, returns default bodycache mode.
function isDynamicPage( $source ) 
{
      // preparation
    if (! is_array($source)) { 
        $lines = explode("\n", $source);
    } else {
        $lines = & $source;
    }

      // check dynamic plugin is included or not.
    while (! empty($lines)) {
        $line = array_shift($lines);
		if( preg_match('/^#([^\(\{]+)(?:\(([^\r]*)\))?(\{*)/', $line, $matches) ) {
            $matche1 = strtolower($matches[1]);
            if ( $matche1 == 'bodycache' && $matches[2] ) { // #bodycache(xxxx)
                $matche2 = strtolower( $matches[2] );
                if        ( 'disable' == $matche2 ) {
                    return true;
                } else if ( 'enable'  == $matche2 ) {
                    return false;
                }
            }
            if ( isDynamicPlugin( $matche1 ) ) {
                return true;
            }
        }
    }

    global $enable_bodycache_default;
    return ! $enable_bodycache_default;
}

function bodycache_serialize_footnote()
{
    global $foot_explain;

    $serialized = "Footnotes: " . count($foot_explain) . "\n";
    foreach( $foot_explain as $key => $elm ) {
        $serialized .= $key . ":" . bodycache_pickle($elm) . "\n";
    }
    return $serialized;
}


function bodycache_save_cache( $page_id, $body )
{
    global $bodycache_rev, $bodycache_db_signature;
    global $title, $page;

    $fb = fopen( get_cachename($page_id), "wb" );
    fwrite($fb, $bodycache_db_signature . "\n");
    fwrite($fb, "Rev: " . $bodycache_rev . "\n");
    fwrite($fb, "Title: " . $title . "\n");
    fwrite($fb, "Page: " . $page . "\n");
    fwrite($fb, bodycache_serialize_footnote() );
    fwrite($fb, $body);
    fclose( $fb );
}

//! parse cache. if cache is obsolete, generate it from default process.
function bodycache_read_cache( $page_id )
{
    global $bodycache_rev, $bodycache_db_signature;
    global $foot_explain;
    global $bodycache_status;
    global $title, $page;

    $body = "";
    $fp = fopen( get_cachename($page_id), "rb");

    while( true ) {

          // signature check.
        if( fgets( $fp ) != $bodycache_db_signature . "\n" ) { 
            bcdump("bad signature:" ); 
            break;
        }

          // revision check.
        if( ! sscanf( fgets($fp), "Rev: %d", $rev) && $rev != $bodycache_rev && $rev != 1 ) { 
            bcdump("bad rev." . $rev );
            break;
        }
        
        if( $rev != 1 ) {
          if( ! sscanf( fgets($fp), "Title: %[^\n]", $title) ) { 
              bcdump("bad title.");
              break;
          }
          if( ! sscanf( fgets($fp), "Page: %[^\n]", $page) ) { 
              bcdump("bad page.");
              break;
          }
        }
        
          // read footnote.
        if( ! sscanf( fgets($fp), "Footnotes: %d", $nnote) ) { 
            bcdump("bad footnote.");
            break;
        }
        for ( $i=0; $i<$nnote; $i++) {
            if( ! sscanf( fgets($fp), "%d:%s\n", $idx, $encoded ) ) { continue; }
            $foot_explain[$idx] = bodycache_unpickle( $encoded );
        }

          // read body.
        do {
            $tmp  = fread( $fp, 8192 );
            $body .= $tmp;
        } while ( strlen($tmp) != 0 );
        fclose( $fp );

          // successful return.
        $bodycache_status = 'cached';
        return '<!-- read from bodycache. Rev: ' . $bodycache_rev . ' -->' . $body;
    }

      // something is bad. re-generate it.
    return bodycache_default_process($page_id); 
}


function bodycache_default_process( $page )
{
    global $bodycache_rev;
    global $bodycache_status;

    $source = get_source($page);
    $body = convert_html($source);

    if ( ! isDynamicPage($source) ) {
        bcdump("static page...");
        bodycache_save_cache( $page, $body );
        $bodycache_status = 'generated';
        return '<!-- bodycache created. Rev: ' . $bodycache_rev . ' -->' . $body;
    }

    bcdump("dynamic page...");
    $bodycache_status = 'skipped';
    return '<!-- bodycache not created (dynamic page). Rev:' . $bodycache_rev . ' -->' . $body;
}

function is_cache($page, $clearcache = FALSE)
{
	if ($clearcache) clearstatcache();
	return file_exists(get_cachename($page));
}

function default_renderer( $page )
{
    return convert_html(get_source($page));
}

function get_cache($page)
{
    global $bodycache_rev;

    if ( ! is_page($page) ) { // page not exists.
        return default_renderer($page);
    }
    if ( ! is_cache($page) ) { // cache not exists.
        return bodycache_default_process($page);
    }        
    if ( get_filetime($page) > get_cachetime($page) ) { // cache is obsolete.
        return bodycache_default_process($page);
    }
    
    return bodycache_read_cache($page);
}

// Get last-modified filetime of the cache
function get_cachetime($page)
{
	return is_cache($page) ? filemtime(get_cachename($page)) - LOCALZONE : 0;
}

// Get physical file name of the cache
function get_cachename($page)
{
	return CACHE_DIR . encode($page) . '.body';
}

function render_body( $page )
{
      // $body = default_renderer( $page );
    $body = get_cache( $page );
    return $body;
}

// Last-Modified header
function header_lastmod_cache($page = NULL)
{
	global $lastmod;

	if ( is_cache($page) && get_cachetime($page) != 0 ) {
		pkwk_headers_sent();
		header('Last-Modified: ' .
			date('D, d M Y H:i:s', get_cachetime($page)) . ' GMT');
	} else if ( is_page($page) ) {
		header('Last-Modified: ' .
			date('D, d M Y H:i:s', UTIME) . ' GMT');
	}
}
